/* OprofileDb - An Oprofile sample file database wrapper.
 * Written by Keith Seitz  <keiths@redhat.com>
 *
 * (c) 2003 Red Hat, Inc.
 *
 * This program is open source software licensed under the IBM
 * Common Public License ver. 1
 */
#include <iostream>
#include "oprofiledb.h"
#include "stable.h"
#include "sample.h"

#include "op_sample_file.h"

OprofileDb::OprofileDb (const char *filename)
  : _filename (strdup (filename)), _tree (NULL), _symbol_table (NULL)
{
}

OprofileDb::~OprofileDb ()
{
  _close_db ();
  samples_t::iterator i = _samples.begin ();
  while (i != _samples.end ())
    {
      delete SAMPLE (*i);
      ++i;
    }
  _samples.clear ();
  free (_filename);
}

void
OprofileDb::_open_db (void)
{
  if (_tree == NULL)
    {
      int rc;
      char *err_msg;

      _tree = new samples_odb_t;
      rc = odb_open (_tree, _filename, ODB_RDONLY, sizeof (opd_header), &err_msg);
      if (rc != EXIT_SUCCESS)
	{
	  // This shouldn't happen, but let's at least print something out.
	  std::cerr << "Error opening oprofile database: " << err_msg << std::endl;
	  free (err_msg);
	}
    }
}

void
OprofileDb::_close_db (void)
{
  if (_tree != NULL)
    {
      odb_close (_tree);
      delete _tree;
    }

  _tree = NULL;
}

const OprofileDb::samples_t &
OprofileDb::get_samples (SymbolTable *stable)
{
  _open_db ();
  _symbol_table = stable;
  samples_odb_travel (_tree, 0, ~0, _get_samples_callback, this);
  _symbol_table = NULL;
  return _samples;
}

// This is a callback from oprofile when traveling the samples in the sample file.
void
OprofileDb::_get_samples_callback (odb_key_t key, odb_value_t info, void *data)
{
  OprofileDb *odb = (OprofileDb *) data;

  Symbol *symbol;
  bfd_vma real_addr;
  if (odb->_symbol_table != NULL)
    symbol = odb->_symbol_table->lookup_vma ((bfd_vma) key, real_addr);

  // Oprofile can have multiple samples for the same VMA, so look in the
  // our map/database and see if the given VMA exists. If it does not exist,
  // add a new Sample. If it does exist, just increment the count of the Sample
  // by INFO.
  samples_t::iterator i = odb->_samples.find ((bfd_vma) key);
  if (i == odb->_samples.end ())
    {
      // new sample
      Sample *s = new Sample (real_addr, symbol, info);
      odb->_samples.insert (sample_t (key, s));
    }
  else
    {
      // existing sample
      SAMPLE (*i)->incr_count (info);
    }
}

const struct opd_header *
OprofileDb::get_header (void)
{
  _open_db ();
  struct opd_header *header = static_cast<struct opd_header *> (_tree->base_memory);
  return header;
}
