/* Opxml a (really) simple program to output XML descriptions of
   oprofile sample files. This program exists as a bridge between
   GPL'd software (oprofile and BFD) and CPL'd software (Eclipse). */

#include <stdlib.h>
#include <iostream>
#include <string>
#include <getopt.h>
#include <errno.h>
#include <string.h>

#include <op_events.h>
#include <op_config.h>
#include <op_sample_file.h>

#include "stable.h"
#include "sample.h"
#include "oprofiledb.h"

// Enum describing info options
enum options
{
   NUM_COUNTERS = 0,
   EVENT_LIST,
   DEFAULTS,
   CHECK_EVENT,
   LIMITS,
   DEMANGLE_FILENAME,
   DEBUG_INFO,
   HEADER,
   SAMPLES
};

// Strings of options
static const char *options[] =
{
  "num-counters",
  "event-list",
  "defaults",
  "check-events",
  "limits",
  "demangle-filename",
  "debug-info",
  "header",
  "samples",
  0
};

// Help for options
struct help
{
  const char *arg;
  const char *desc;
};

static const struct help args_help[] =
{
  {"\t\t\t", "number of supported counters"},
  {"[CTR]\t\t", "event list for counter (default: first counter)"},
  {"\t\t\t", "get default size for BUFFER, NOTE, HASH"},
  {"CTR EVENT UMASK\t", "check counter/event validity"},
  {"\t\t\t\t", "get oprofile limits"},
  {"FILE\t\t", "demangle sample file name"},
  {"EXE VMA\t\t", "show debug info for address VMA in EXE_NAME"},
  {"FILE\t\t\t", "show header info for sample FILE"},
  {"FILE\t\t\t", "get samples for given sample FILE"}
};

// Local functions
static op_cpu str_to_op_cpu (const char *const cpu_str);
static void print_usage (const char *const argv0);
static int get_option_index (const char *options[], const char *arg);
static void output_cpu_type (const char *name, op_cpu cpu_type);
static void output_unit_mask (struct op_event *event);
static void output_dtd_header (void);
static void wrong_num_arguments (int argc, char *argv[], const char *opts);
static std::string escape (std::string &string);

// Info handlers
static int num_counters (op_cpu cpu_type, int argc, char *argv[]);
static int event_list (op_cpu cpu_type, int argc, char *argv[]);
static int defaults (op_cpu cpu_type, int argc, char *argv[]);
static int check_events (op_cpu cpu_type, int argc, char *argv[]);
static int limits (op_cpu cpu_type, int argc, char *argv[]);
static int demangle_filename (op_cpu cpu_type, int argc, char *argv[]);
static int debug_info (op_cpu cpu_type, int argc, char *argv[]);
static int header (op_cpu cpu_type, int argc, char *argv[]);
static int samples (op_cpu cpu_type, int argc, char *argv[]);

static void
wrong_num_arguments (int argc, char *argv[], const char *opts)
{
  std::cerr << "wrong # args: should be \"";
  for (int i = 0; i < argc; ++i)
    std::cerr << argv[i] << " ";

  std::cerr << opts << "\"" << std::endl;
  exit (EXIT_FAILURE);
}

// Converts the argument into its corresponding option index
static int
get_option_index (const char *options[], const char *arg)
{
  const char *option;

  int i = 0;
  for (option = options[0]; option != NULL; option = options[++i])
    {
      if (strncmp (option, arg, strlen (arg)) == 0)
	return i;
    }

  return -1;
}

// Prints a small help message
static void
print_help (const char *const argv0)
{
  std::cerr << "Use '" << argv0 << " --help' for a complete list of options."
	    << std::endl;
}

// Prints the usage of this program
static void
print_usage (const char *const argv0)
{
  std::cerr << argv0 << ": usage: " << argv0 << " [OPTION] INFO [TYPE_OPTION]" << std::endl;
  std::cerr << "Supply information about Oprofile and its sample files." << std::endl << std::endl;
  std::cerr << "Options:" << std::endl << std::endl;
  std::cerr << "--cpu|-c CPU_TYPE\t\t\t   cpu type (current: "
	    << op_get_cpu_type_str (op_get_cpu_type ()) << ")" << std::endl;

  std::cerr << std::endl << "Types of INFO:" << std::endl << std::endl;

  int i;
  for (i = 0; i < (sizeof (args_help) / sizeof (args_help[0])); ++i)
    {
      std::cerr << options[i] << " " << args_help[i].arg << "   "
		<< args_help[i].desc << std::endl;
    } 
}

// Convenience function to print out the line:
// <NAME cpu_type="PIII"/>
static void
output_cpu_type (const char *name, op_cpu cpu_type)
{
  output_dtd_header ();
  std::cout << "<" << name << " cpu_type=\"" << op_get_cpu_type_str (cpu_type) << "\">"
	    << std::endl;
}

static void
output_dtd_header (void)
{
  std::cout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
}

int
main (int argc, char *argv[])
{
  const char *argv0 = argv[0];
  op_cpu cpu_type = op_get_cpu_type ();

  if (argc < 2)
    {
      print_help (argv0);
      exit (EXIT_FAILURE);
    }

  static struct option long_options[] =
  {
    {"cpu-type", required_argument, 0, 'c'},
    {"help", no_argument, 0, 'h'},
    {0, 0, 0, 0}
  };

  int index;
  while (true)
    {
      char c;
      c = getopt_long_only (argc, argv, "", long_options, &index);
      if (c == -1)
	break;

      if (c == 0 && long_options[index].flag == 0)
	c = long_options[index].val;

      switch (c)
	{
	case 'c':
	  cpu_type = str_to_op_cpu (optarg);
	  if (cpu_type == CPU_NO_GOOD)
	    {
	      std::cerr << argv0 << ": cpu \"" << optarg << "\" not recognized" << std::endl;
	      exit (EXIT_FAILURE);
	    }
	  break;

	case 'h':
	  print_usage (argv0);
	  exit (EXIT_SUCCESS);
	  break;

	case '?':
	  print_help (argv0);
	  exit (EXIT_FAILURE);
	}
    }

  // Check that we have a valid cpu type. It can be invalid (here) because
  // the Oprofile module not loaded/running. Oprofile library will output error.
  if (cpu_type == CPU_NO_GOOD)
    exit (EXIT_FAILURE);

  argc -= optind;
  argv += optind;

  if (argc == 0)
    {
      print_help (argv0);
      exit (EXIT_FAILURE);
    }

  int rc;
  index = get_option_index (options, argv[0]);
    {
      switch ((enum options) index)
	{
	case NUM_COUNTERS:
	  rc = num_counters (cpu_type, argc, argv);
	  break;

	case EVENT_LIST:
	  rc = event_list (cpu_type, argc, argv);
	  break;

	case DEFAULTS:
	  rc = defaults (cpu_type, argc, argv);
	  break;

	case CHECK_EVENT:
	  rc = check_events (cpu_type, argc, argv);
	  break;

	case LIMITS:
	  rc = limits (cpu_type, argc, argv);
	  break;

	case DEMANGLE_FILENAME:
	  rc = demangle_filename (cpu_type, argc, argv);
	  break;

	case DEBUG_INFO:
	  rc = debug_info (cpu_type, argc, argv);
	  break;

	case HEADER:
	  rc = header (cpu_type, argc, argv);
	  break;

	case SAMPLES:
	  rc = samples(cpu_type, argc, argv);
	  break;

	default:
	  std::cerr <<  argv0 << ": unknown option \"" << argv[0] << "\"" << std::endl;
	  print_help (argv0);
	  rc = EXIT_FAILURE;
	}
    }

  return rc;
}

// Converts the given string into a corresponding op_cpu (CPU_NO_GOOD if invalid)
static op_cpu
str_to_op_cpu (const char *const cpu_str)
{
  int i;
  for (i = 0; i < MAX_CPU_TYPE; ++i)
    {
      if (strcmp (op_get_cpu_type_str ((op_cpu) i), cpu_str) == 0)
	return (op_cpu) i;
    }

  return CPU_NO_GOOD;
}

/* Get an integer value from the argument. Only does ints > 0.
   Returns the integer or -1. */
static int
get_integer (const char *arg)
{
  errno = 0;
  int integer = strtol (arg, NULL, 10);
  if (errno != 0)
    return -1;
  
  return integer;
}

struct escape_chars
{
  const char *bad;
  const char *good;
};

static struct escape_chars escape_characters[] =
{
  {"&",  "&amp;"},
  {"<",  "&lt;"},
  {">",  "&gt;"},
  {"\'", "&apos;"},
  {"\"", "&quot;"}  
};

static std::string
escape (std::string &string)
{
  for (int i = 0; i < sizeof (escape_characters) / sizeof (escape_characters[0]); ++i)
    {
      size_t pos = string.find (escape_characters[i].bad);
      while (pos != std::string::npos)
	{
	  string.replace (pos, 1, escape_characters[i].good);
	  pos = string.find (escape_characters[i].bad, pos + 1);
	}
    }

  return string;
}

/* Print the number of counters for the given cpu type
 *
 * Input: <none>
 * Output:
 * <num-counters cpu_type="PIII">
 *    <counters>2</counters>
 * </num-counters>
 */
static int
num_counters (op_cpu cpu_type, int argc, char *argv[])
{
  output_cpu_type ("num-counters", cpu_type);
  std::cout << "\t<counters>" << op_get_nr_counters (cpu_type) << "</counters>" << std::endl;
  std::cout << "</num-counters>" << std::endl;
  return EXIT_SUCCESS;
}

/* Print out the valid events for the given cpu_type and
 * counter number in argv.
 *
 * Input: [counter-number]
 * Output:
 * <event-list cpu_type="PIII" counter="1">
 *    <event>
 *       <name>EVENT1</name>
 *       <description>A description of EVENT1</description>
 *       <value>1</value>
 *       <minimum>30000</minimum>
 *       <unit-mask>
 *          <default>0</default>
 *          <type>mandatory</type>
 *          <mask>
 *             <value>1</value>
 *             <description>Description of mask value 1</description>
 *          </mask>
 *          ...
 *          <mask>
 *             <value>N</value>
 *             <description>Description of mask value N</description>
 *          </mask>
 *       </unit-mask>
 *    </event>
 *    ...
 *    <event>
 *       <name>EVENTN</name>
 *       <description>A description of EVENTN</description>
 *       <value>N</value>
 *    <event>
 * </event-list>
 */
static int
event_list (op_cpu cpu_type, int argc, char *argv[])
{
  int counter = 0; // default to first counter
  if (argc == 2)
    {
      counter = get_integer (argv[1]);
      if (counter < 0)
	{
	  std::cerr << "invalid counter \"" << argv[1] << "\": "
		    << strerror (errno) << std::endl;
	  return EXIT_FAILURE;
	}
    }

  if (counter >= op_get_nr_counters (cpu_type))
    {
      std::cerr << "counter must be between 0 and " << (op_get_nr_counters (cpu_type) - 1)
		<< std::endl;
      return EXIT_FAILURE;
    }

  output_cpu_type ("event-list", cpu_type);

  struct list_head *events, *p;
  events = op_events (cpu_type);
  list_for_each (p, events)
    {
      struct op_event *event = list_entry (p, struct op_event, event_next);
      if (/*event->counter_mask == CTR_ALL ||*/ (event->counter_mask & (1 << counter)) > 0)
	{
	  std::cout << "\t<event>" << std::endl;
	  std::string str(event->name);
	  std::cout << "\t\t<name>" << escape (str) << "</name>" << std::endl;
	  str.assign (event->desc);
	  std::cout << "\t\t<description>" << escape (str) << "</description>"
		    << std::endl;
	  std::cout << "\t\t<value>" << (int) event->val << "</value>" << std::endl;
          std::cout << "\t\t<minimum>" << (int) event->min_count << "</minimum>"
                    << std::endl;
	  output_unit_mask (event);
	  std::cout << "\t</event>" << std::endl;
	}
    }
  std::cout << "</event-list>" << std::endl;

  return EXIT_SUCCESS;
}

static void
output_unit_mask (struct op_event *event)
{
  const struct op_unit_mask *umask = event->unit;
  const char *type;
  switch (umask->unit_type_mask)
    {
    case utm_mandatory:
      type = "mandatory";    break;
    case utm_exclusive:
      type = "exclusive";    break;
    case utm_bitmask:
      type = "bitmask";      break;
    }

  std::cout << "\t\t<unit-mask>" << std::endl;
  std::cout << "\t\t\t<type>" << type << "</type>" << std::endl;
  std::cout << "\t\t\t<default>" << umask->default_mask << "</default>" << std::endl;

  for (int i = 0; i < umask->num; ++i)
    {
      std::cout << "\t\t\t<mask>" << std::endl;
      std::cout << "\t\t\t\t<value>" << umask->um[i].value << "</value>" << std::endl;
      std::string str(umask->um[i].desc);
      std::cout << "\t\t\t\t<description>" << escape (str) << "</description>"
		<< std::endl;
      std::cout << "\t\t\t</mask>" << std::endl;
    }
  std::cout << "\t\t</unit-mask>" << std::endl;
}

/* Print out default information for oprofile.
 *
 * Input: <none>
 * Output:
 * <defaults cpu_type="PIII">
 *    <buffer-size>
 *       <min>1</min>
 *       <max>65535</max>
 *       <default>32767</default>
 *    </buffer-size>
 *    <note-size>
 *       <min>1</min>
 *       <max>65535</max>
 *       <default>32767</default>
 *    </note-size>
 *    <sample-dir>/var/lib/oprofile/samples</sample-dir>
 *    <lock-file>/var/lib/oprofile/lock</lock-file>
 *    <log-file>/var/lib/oprofile/oprofiled.log</log-file>
 *    <dump-status/var/lib/oprofile/complete_dump</dump-status>
 * </defaults>
 * 
 * Note: Not all defaults supported on all versions of oprofile!
 */
static int
defaults (op_cpu cpu_type, int argc, char *argv[])
{
  output_cpu_type ("defaults", cpu_type);
#if defined(OP_MIN_BUF_SIZE)
  std::cout << "\t<buffer-size>" << std::endl;
  std::cout << "\t\t<min>" << OP_MIN_BUF_SIZE << "</min>" << std::end;
  std::cout << "\t\t<max>" << OP_MAX_BUF_SIZE << "</max>" << std::end;
  std::cout << "\t\t<default>" << OP_DEFAULT_BUF_SIZE << "</default>" << std::endl;
  std::cout << "\t</buffer-size>" << std::endl;
#endif
#if defined(OP_MIN_NOTE_TABLE_SIZE)
  std::cout << "\t<note-size>" << std::endl;
  std::cout << "\t\t<min>" << OP_MIN_NOTE_TABLE_SIZE << "</min>" << std::end;
  std::cout << "\t\t<max>" << OP_MAX_NOTE_TABLE_SIZE << "</max>" << std::end;
  std::cout << "\t\t<default>" << OP_DEFAULT_NOTE_SIZE << "</default>" << std::endl;
  std::cout << "\t</note-size>" << std::endl;
#endif

  std::string str(OP_SAMPLES_DIR);
  std::cout << "\t<sample-dir>" << escape (str) << "</sample-dir>" << std::endl;
  str.assign (OP_LOCK_FILE);
  std::cout << "\t<lock-file>" << escape (str) << "</lock-file>" << std::endl;
  str.assign (OP_LOG_FILE);
  std::cout << "\t<log-file>" << escape (str) << "</log-file>" << std::endl;
  str.assign (OP_DUMP_STATUS);
  std::cout << "\t<dump-status>" << escape (str) << "</dump-status>" << std::endl;

  std::cout << "</defaults>" << std::endl;
 
  return EXIT_SUCCESS;
}

/* Check whether the given counter/event/umask info is valid.
 *
 * Input: counter-number event-number unit-mask
 * Output:
 * <check-events>
 *    <result>{ok | invalid-{event,counter,um}}</result>
 * </check-events>
 *
 * Note: output is a BITMASK of errors. Expect multiple "result" fields on error.
 */
static int
check_events (op_cpu cpu_type, int argc, char *argv[])
{
  static const char *errors[3] = {"invalid-event", "invalid-um", "invalid-counter"};

  if (argc != 4)
    wrong_num_arguments (1, argv, "counter event umask");

  int counter = get_integer (argv[1]);
  if (counter < 0)
    {
      std::cerr << "invalid counter \"" << argv[2] << "\": "
		<< strerror (errno) << std::endl;
      return EXIT_FAILURE;
    }

  if (counter >= op_get_nr_counters (cpu_type))
    {
      std::cerr << "counter must not be greater than " << (op_get_nr_counters (cpu_type) - 1)
		<< std::endl;
      return EXIT_FAILURE;
    }

  int event = get_integer (argv[2]);
  if (event < 0)
    {
      std::cerr << "invalid event \"" << argv[1] << "\": "
		<< strerror (errno) << std::endl;
      return EXIT_FAILURE;
    }

  int umask = get_integer (argv[3]);
  if (umask < 0)
    {
      std::cerr << "invalid unit mask \"" << argv[2] << "\": "
		<< strerror (errno) << std::endl;
      return EXIT_FAILURE;
    }

  int result = op_check_events (counter, event, umask, cpu_type);

  output_cpu_type ("check-events", cpu_type);

  if (result == (int) OP_OK_EVENT)
    std::cout << "\t<result>ok</result>" << std::endl;
  else
    {
      for (int i = 0; i < sizeof(errors) / sizeof(errors[0]); ++i)
	{
	  if ((result & (1 << i)) > 0)
	    std::cout << "\t<result>" << errors[i] << "</result>" << std::endl;
	}
    }

  std::cout << "</check-events>" << std::endl;
  return EXIT_SUCCESS;
}

/* Print out the limits.
 *
 * Output:
 * ???
 */
static int
limits (op_cpu cpu_type, int argc, char *argv[])
{
  return EXIT_SUCCESS;
}

static void
demangle_sample_filename (const char *sample_file, std::string &demangled_name)
{
  demangled_name = sample_file;
  size_t pos = demangled_name.find_first_of (OPD_MANGLE_CHAR);
  if (pos != std::string::npos)
    {
      demangled_name.erase (0, pos);
      replace (demangled_name.begin (), demangled_name.end (), OPD_MANGLE_CHAR, '/');
    }
  pos = demangled_name.find_last_of ('#');
  if (pos != std::string::npos)
    demangled_name.erase (pos, std::string::npos);
}

/* Print out the demangled filename for the given sample file in argv.
 *
 * Input: sampleFile
 * Output:
 * <demangle-filename>
 *    <filename>/home/keiths/foo</filename>
 * </demangle-filename> 
 */
static int
demangle_filename (op_cpu cpu_type, int argc, char *argv[])
{
  if (argc != 2)
    wrong_num_arguments (1, argv, "filename");

  std::string demangled_name;
  demangle_sample_filename (argv[1], demangled_name);
  output_cpu_type ("demangle-filename", cpu_type);
  std::cout << "\t<filename>" << escape (demangled_name) << "</filename>" << std::endl;
  std::cout << "</demangle-filename>" << std::endl;

  return EXIT_SUCCESS;
}

/* Print out the debug info for the given executable and VMA in argv.
 *
 * Input: executable vma
 * Output:
 * <debug-info cpu_type="PIII">
 *    <source-filename>/home/keiths/foo.c</source-filename>
 *    <function>bar</function>
 *    <line>12</line>
 * </debug-info>
 */
static int
debug_info (op_cpu cpu_type, int argc, char *argv[])
{
  if (argc != 3)
    wrong_num_arguments (1, argv, "executable vma");

  const char *end;
  const char *executable = argv[1];
  bfd_vma vma = bfd_scan_vma (argv[2], &end, 0);

  // Read in symbols
  SymbolTable st (executable);
  st.read_symbols ();

  unsigned int line;
  const char *func, *file;
  st.get_debug_info (vma, func, file, line);

  std::string str;
  output_cpu_type ("debug-info", cpu_type);
  if (func != NULL)
    {
      str.assign (func);
      std::cout << "\t<function>" << escape (str) << "</function>" << std::endl;
    }

  if (file != NULL)
    {
      str.assign (file);
      std::cout << "\t<source-file>" << escape (str) << "</source-file>" << std::endl;
    }

  if (line != 0)
    std::cout << "\t<line>" << line << "</line>" << std::endl;

  std::cout << "</debug-info>" << std::endl;

  return EXIT_SUCCESS;
}

/* Print out header information for the sample file in argv.
 *
 * Input: sample file
 * Output:
 * <header cpu_type="PIII">
 *    <cpu_type>PIII</cpu_type>
 *    <counter>0</counter>
 *    <event>CPU_CLK_UNHALTED</event>
 *    <count>65535</count>
 *    <unit-mask>0</unit-mask>
 *    <cpu-speed>0</cpu-speed>
 *    <separate-lib>0</separate-lib>
 *    <separate-kernel>0</separate-kernel>
 * </header>
 */
static int
header (op_cpu cpu_type, int argc, char *argv[])
{
  if (argc != 2)
    wrong_num_arguments (1, argv, "sample_file");

  OprofileDb db (argv[1]);
  const struct opd_header *header = db.get_header ();

  output_cpu_type ("header", cpu_type);

  std::cout << "\t<cpu_type>" << op_get_cpu_type_str ((op_cpu) header->cpu_type)
	    << "</cpu_type>" << std::endl;
  std::cout << "\t<counter>" << header->ctr << "</counter>" << std::endl;
  std::cout << "\t<event>" << header->ctr_event << "</event>" << std::endl;
  std::cout << "\t<count>" << header->ctr_count << "</count>" << std::endl;
  std::cout << "\t<unit-mask>" << header->ctr_um << "</unit-mask>" << std::endl;
  std::cout << "\t<cpu-speed>" << header->cpu_speed << "</cpu-speed>" << std::endl;
  std::cout << "\t<separate-lib>" << header->separate_lib_samples << "</separate-lib>"
	    << std::endl;
  std::cout << "\t<separate-kernel>" << header->separate_kernel_samples
	    << "</separate-kernel>" << std::endl;

  std::cout << "</header>" << std::endl;

  return EXIT_SUCCESS;
}

/* Print out the samples associated with the given sample file
 *
 * Input: sampleFile
 * Output:
 * <sample>
 *    <addr>0x12345678</addr>
 *    <count>12</count>
 *    <symbol>
 *       <addr>0x1234568</addr>
 *       <name>foo</name>
 *    </symbol>
 * </sample>
 * <sample>
 * ...
 */
static int
samples (op_cpu cpu_type, int argc, char *argv[])
{
  if (argc != 2)
    wrong_num_arguments (1, argv, "sampleFile");

  const char *sample_file = argv[1];
  std::string executable;
  demangle_sample_filename (sample_file, executable);

  // Read symbols for the given executable
  SymbolTable st (executable.c_str ());
  st.read_symbols ();

  // Get the samples from the sample file
  OprofileDb db (sample_file);
  OprofileDb::samples_t samples = db.get_samples (&st);

  output_cpu_type ("samples", cpu_type);

  // Loop through and output the samples
  OprofileDb::samples_t::iterator i = samples.begin ();
  while (i != samples.end ())
    {
      Sample *smpl = OprofileDb::SAMPLE (*i);

      char buf[257];
      sprintf_vma (buf, smpl->vma ());

      std::cout << "\t<sample>" << std::endl;
      std::cout << "\t\t<addr>0x" << buf << "</addr>" << std::endl;
      std::cout << "\t\t<count>" << smpl->count () << "</count>" << std::endl;

      if (smpl->has_symbol ())
	{
	  std::cout << "\t\t<symbol>" << std::endl;
	  std::string str (smpl->demangled_name ());
	  std::cout << "\t\t\t<name>" << escape (str) << "</name>"
		    << std::endl;
	  sprintf_vma (buf, smpl->symbol ()->start ());
	  std::cout << "\t\t\t<addr>0x" << buf << "</addr>" << std::endl;
	  std::cout << "\t\t</symbol>" << std::endl;
	}

      std::cout << "\t</sample>" << std::endl;
      ++i;
    }

  std::cout << "</samples>" << std::endl;
  return EXIT_SUCCESS;
}

