/* SymbolTable - A symbol table class
 * Written by Keith Seitz  <keiths@redhat.com>
 *
 * (c) 2003 Red Hat, Inc.
 *
 * This program is open source software licensed under the IBM
 * Common Public License ver. 1
 */
#include <stdlib.h>
#include <string.h>

#include <algorithm>
#include <vector>

#include "stable.h"
#include "symbol.h"
#include "sample.h"

#include "elf/common.h"
#include "elf/internal.h"

static bool ltvma (const Symbol *a, const Symbol *b);

// Stolen from oprofile
char const *SymbolTable::_boring_symbols[] = {
  "gcc2_compiled.",
  "_init"
};

// Helper function for sorting symbols by VMA
static bool
compare_symbol_with_vma (bfd_vma vma, const Symbol *a)
{
  return (a->end () > vma);
}

SymbolTable::SymbolTable(const char *file)
  : _filename (strdup (file)), _abfd (NULL), _symbol_table (NULL), _cache_symbol (NULL)
{
}

SymbolTable::~SymbolTable ()
{
  if (_symbol_table != NULL)
    free (_symbol_table);

  std::vector<Symbol *>::iterator i = _symbols.begin ();
  while (i != _symbols.end ())
    {
      delete (*i);
      //_symbols.erase (i);
      ++i;
    }

  _symbols.clear ();
  _close_bfd ();
  free (_filename);
}

// If this returns NULL, then the VMA is not in any range of
// msymbols. This is can apparently happen. op_time and friends
// ignore these samples.
#define PRINT 0
Symbol *
SymbolTable::lookup_vma (bfd_vma vma, bfd_vma &real_vma)
{
  real_vma = vma + _start_vma;
  return (lookup_vma (real_vma));
}

Symbol *
SymbolTable::lookup_vma (bfd_vma real_vma)
{
  if (_cache_symbol != NULL && _cache_symbol->contains (real_vma))
    return _cache_symbol;

  std::vector<Symbol *>::iterator i;
  i = upper_bound (_symbols.begin (), _symbols.end (),
		   real_vma, compare_symbol_with_vma);
  if (i != _symbols.end () && (*i)->contains (real_vma))
    {
      _cache_symbol = *i;
      return *i;
    }

  return NULL;
}

bool
SymbolTable::read_symbols ()
{
  if (_open_bfd ())
    {
      long storage_needed = bfd_get_symtab_upper_bound (_abfd);
      if (storage_needed > 0)
	{
	  _symbol_table = (asymbol **) malloc (storage_needed);
	  long num_symbols = bfd_canonicalize_symtab (_abfd, _symbol_table);
	  for (int i = 0; i < num_symbols; ++i)
	    {
	      if (_interesting_symbol (_symbol_table[i]))
		_symbols.push_back (new Symbol (_symbol_table[i]));
	    }

	  if (_symbols.size() > 0)
	    {
	      // Sort in order of increasing vma and eliminate duplicates
	      stable_sort (_symbols.begin (), _symbols.end (), ltvma);

	      // Eliminate duplicates
	      for (int i = 0; i < _symbols.size () - 1; ++i)
		{
		  if (_symbols[i]->start () == _symbols[i+1]->start ())
		    {
		      int erase, keep;

		      // Opt to keep FUNCTIONs, first come, first kept
		      if (_symbols[i]->flags () & BSF_FUNCTION)
			erase = i + 1;
		      else if (_symbols[i+1]->flags () & BSF_FUNCTION)
			erase = i;
		      else // Don't know. Keep first.
			erase = i + 1;

		      delete *(_symbols.begin () + erase);
		      _symbols.erase (_symbols.begin () + erase);
		      --i;
		    }
		}
	      
	      // Fill in end addresses
	      for (int i = 0; i <= _symbols.size () - 1; ++i)
		{
		  asymbol *this_sym, *next_sym;
		  asection *this_sect, *next_sect;

		  this_sym = _symbols[i]->get_asymbol ();
		  this_sect = bfd_get_section (this_sym);
		  if (i < _symbols.size () - 1)
		    {
		      next_sym = _symbols[i + 1]->get_asymbol ();
		      next_sect = bfd_get_section (next_sym);
		    }
		  else
		    next_sect = NULL;

		  if (next_sect != NULL
		      && bfd_get_section (this_sym) == bfd_get_section (next_sym))
		    _symbols[i]->end (_symbols[i + 1]->start ());
		  else
		    {
		      asection *asect = bfd_get_section (this_sym);
		      bfd_vma end = bfd_get_section_vma (_abfd, asect);
		      end += bfd_section_size (_abfd, asect);
		      _symbols[i]->end (end);
		    }
		}

	      return true;
	    }
	}
    }

  return false;
}

bool
SymbolTable::_interesting_symbol (asymbol *sym)
{
  if (!(bfd_get_section_flags (bfd_asymbol_bfd (sym), bfd_get_section (sym)) & SEC_CODE))
    return false;

  const char *name = bfd_asymbol_name (sym);
  if (name == NULL || name[0] == '\0')
    return false;

  // C++ exception stuff
  if (name[0] == '.' && name[1] == 'L')
    return false;

  for (size_t i = 0; i < sizeof (_boring_symbols) / sizeof (_boring_symbols[0]); i++)
    {
      if (strcmp (name, _boring_symbols[i]) == 0)
	return false;
    }

  return true;
}

bool
SymbolTable::_open_bfd (void)
{
  if (_abfd == NULL)
    {
      bfd_init ();
      _abfd = bfd_openr (_filename, NULL);
      if (_abfd != NULL)
	{
	  char **matches;
	  if (bfd_check_format_matches (_abfd, bfd_object, &matches))
	    {
	      // Get the physical load address
	      long phdr_upper_bound;
	      long phdr_count;
	      Elf_Internal_Phdr *phdrs;
	      int ld_hdr;

	      // This seems lame, since it requires intimate knowledge
	      // about the ELF file. (What if the file is not ELF?)
	      phdr_upper_bound =  bfd_get_elf_phdr_upper_bound (_abfd);
	      phdrs = (Elf_Internal_Phdr *) malloc (phdr_upper_bound);
	      phdr_count = bfd_get_elf_phdrs (_abfd, phdrs);

	      for (ld_hdr = 0; ld_hdr < phdr_count; ld_hdr++)
		{
		  if ((phdrs[ld_hdr].p_type == PT_LOAD)
		      && (phdrs[ld_hdr].p_flags & (PF_R | PF_X)))
		    {
		      _start_vma = phdrs[ld_hdr].p_paddr;
		      break;
		    }
		}
	      free (phdrs);
	    }
	}
    }

  return (_abfd != NULL);
}

void
SymbolTable::_close_bfd (void)
{
  if (_abfd != NULL)
    bfd_close (_abfd);

  _abfd = NULL;
}

bool
SymbolTable::get_debug_info (bfd_vma vma, const char * &function,
			     const char * &source_file, unsigned int &line)
{
  function = NULL;
  source_file = NULL;
  line = 0;

  Symbol *symbol = lookup_vma (vma);
  if (symbol != NULL)
    {
      asection *asection = symbol->section ();

      bfd_vma pc = vma - bfd_get_section_vma (_abfd, asection);
      return bfd_find_nearest_line (_abfd, asection, _symbol_table, pc, 
				    &source_file, &function, &line);
    }

  return false;
}

void
SymbolTable::dump_table (void)
{
  printf ("%8s\t%8s\t%s\n", "start", "end", "name");

  std::vector<Symbol *>::iterator i;
  for (i = _symbols.begin (); i != _symbols.end (); ++i)
    {
      Symbol *sym = *i;;
      printf("%8x\t%8x\t%s\n", sym->start (), sym->end (),
	     sym->name ());
    }
}

// Helper function to sort two symbols based on VMA
static bool
ltvma (const Symbol *a, const Symbol *b)
{
  return (a->start () < b->start ());
}

